;------------------------------------------------------------------------------
;		F R E E - D O S ( 9 8 ) X M S - D R I V E R
;------------------------------------------------------------------------------
;
; Improvments by Martin Stroemberg.
; Copyright 2001, 2002, Martin Stroemberg.
;
; Martin can be reached at <ams@ludd.luth.se>. Please start the subject line
; with "FDXMS". If you have a bug report, use bugtrack on the FreeDOS site,
; <http://www.freedos.org>. Make sure to include the FDXMS.SYS version number.
;
;
;------------------------------------------------------------------------------
; Written by Till Gerken for the Free-DOS project.
;
; If you would like to use parts of this driver in one of your projects, please
; check up with me first.
;
; Till can be reached at:	till@tantalo.net (Internet)
;
; For questions concerning Free-DOS, mail the coordinator
; Morgan "Hannibal" Toal <hannibal@iastate.edu>
;
; Comments and bug reports are always appreciated.
;
; Copyright (c) 1995, Till Gerken
;------------------------------------------------------------------------------
; -- IMPLEMENTATION NOTES --
;
; - The Protected Mode handling is very simple, it fits only for the least
;   needs
; - I didn't care about reentrancy. If this thing should be used in
;   Multitasking Environments, well, somebody got to take a look at the code.
; - INT15h, Func. 88h (Extended Memory Size) has been re-implemented to
;   return 0.
; - Function 0Bh (Move Block) uses it's own Protected Mode handling.
;   It doesn't provide Interrupt windows.
; - The code is not very optimised, I just wrote it down for now. Later, when
;   everything is tested, I'm going to see what can be done
; - Some ideas were taken from the original XMS driver written by
;   Mark E. Huss (meh@bis.adp.com), but everything has been completely
;   rewritten, so if there are bugs, they are mine, not his. ;)
;------------------------------------------------------------------------------

%include "fdxms.inc"			; this file contains structures,
					; constants and the like
%include "version.inc"

;------------------------------------------------------------------------------
; 16-bit resident code and data
;------------------------------------------------------------------------------

segment .text
	align	1
	org	0

;------------------------------------------------------------------------------
; device driver header

		dd	-1			; last driver in list
		dw	8000h			; driver flags
		dw	strategy		; pointer to strategy routine
		dw	interrupt		; pointer to interrupt handler
		db	'XMSXXXX0'		; device driver name

;------------------------------------------------------------------------------
; global data

; for 286

; 102bytes (x2)
loadall286_data:
		times 2 dw 0
		dw	0			; msw
		times 8 dw 0
		dw	0			; tr_reg
		dw	0			; flags
		dw	..@to_loadall		; ip
		dw	0			; ldt_reg
		dw	0			; ds_reg
		dw	0			; ss_reg
		dw	0			; cs_reg
		dw	0			; es_reg
		dw	0			; di
		dw	loadall286_save		; si
		dw	0			; bp
		dw	0			; sp
		dw	0			; bx
		dw	0			; dx
		dw	loadall286_tbl_size / 2	; cx
		dw	0			; ax
		; ds	base:000000h limit:ffffh access:93h[P|type:rwa]
		db	0, 0, 0, 93h, 0ffh, 0ffh	; es_desc
		; ds	base:000000h limit:ffffh access:9bh[P|type:xra]
		db	0, 0, 0, 9bh, 0ffh, 0ffh	; cs_desc
		; ds	base:000000h limit:ffffh access:93h[P|type:rwa]
		db	0, 0, 0, 93h, 0ffh, 0ffh	; ss_desc
		; ds	base:000000h limit:ffffh access:93h[P|type:rwa]
		db	0, 0, 0, 93h, 0ffh, 0ffh	; ds_desc
		;
		db	0, 0, 0, 93h, 0ffh, 0ffh	; gdt_desc
		db	0, 0, 0, 93h, 0ffh, 0ffh	; ldt_desc
		db	0, 0, 0, 93h, 0ffh, 0ffh	; idt_desc
		db	0, 0, 0, 93h, 0ffh, 0ffh	; tss_desc
loadall286_save:

; for 386+ (move used protected mode)

; 30bytes
gdt32		dw	gdt_size, dummy, 0
dummy		times 8 db 0
		; base:00000000h limit:0ffffh attrib:09ah[P|S|type:xr]
code16dsc	db	0ffh, 0ffh, 0, 0, 0, 9ah, 0, 0
		; base:00000000h limit:fffffh(*4k) attrib:c92h[G|DB|P|S|type:rw]
core32dsc	db	0ffh, 0ffh, 0, 0, 0, 92h, 0cfh, 0
gdt_size	equ	$ - dummy - 1
code16idx	equ	code16dsc - dummy
core32idx	equ	core32dsc - dummy

; for 286+ (move used bios)

; 48bytes
bios_gdt: ; see int1f-90
		; reserved
		times 8 db 0			; Dummy entries for BIOS.
		; reserved
		times 8 db 0			; Dummy entries for BIOS.
		; ds	base:00000000h limit:fffffh attrib:093h[P|S|type:rwa]
		dw	0ffffh			; Bits 0-15 of source segment length.
bios_src_low	dw	0			; Bits 0-15 of source address.
bios_src_middle	db	0			; Bits 16-23 of source address.
		db	93h			; Source access rights.
		db	0fh			; More type bits and bits 16-19 of source segment length.
bios_src_high	db	0			; Bits 24-31 of source address.
		; es	base:00000000h limit:fffffh attrib:093h[P|S|type:rwa]
		dw	0ffffh			; Bits 0-15 of dest. segment length.
bios_dst_low	dw	0			; Bits 0-15 of dest. address.
bios_dst_middle	db	0			; Bits 16-23 of dest. address.
		db	93h			; Dest. access rights.
		db	0fh			; More type bits and bits 16-19 of dest segment length.
bios_dst_high	db	0			; Bits 24-31 of dest. address.
		; cs
		times 8 db 0			; Dummy entries for BIOS.
		; ss
		times 8 db 0			; Dummy entries for BIOS.

	; for 286
	times loadall286_tbl_size - ($ - loadall286_save) db 0

; ---

old_int1f	dd	0			; old INT1fh vector
old_int2f	dd	0			; old INT2fh vector

request_ptr	dd	0			; pointer to request header

xms_num_handles	dw	32			; number of available handles

flags		db	0

hma_used	db	0			; set if HMA is used
hma_min		dw	0			; minimal space in HMA that
						; has to be requested
a20_locks	dw	0			; internal A20 lock count

;------------------------------------------------------------------------------
; strategy routine. is called by DOS to initialize the driver once.
; only thing to be done here is to store the address of the device driver
; request block.
; In:	ES:BX - address of request header
; Out:	nothing

strategy:
	mov	[cs:request_ptr + 2], es	; store segment addr
	mov	[cs:request_ptr], bx		; store offset addr
	retf					; far return here!

;------------------------------------------------------------------------------
; interrupt routine. called by DOS right after the strategy routine to
; process the incoming job. also used to initialize the driver.

interrupt:
	pushf
	push	di
	push	es

	push	cs
	pop	ds

	les	di, [request_ptr]		; load address of request header

	cmp	byte [es:di + request_hdr.cmd], CMD_INIT ; do we have to initialize?
	jne	.done
	test	byte [flags], FLAG_INITIALISED	; Have we initialized
						; already?
	jnz	.done
	call	initialize			; No, do it now!
.done:

	lds	si, [request_ptr]		; return this to DOS

	pop	es
	pop	di
	popf
	retf					; far return here!

;------------------------------------------------------------------------------
; enables the A20 address line
;	AL - destroyed.

enable_a20:
	; for 286+
	xor	al, al
	out	0f2h, al	; clear a20mask (286+)
	mov	al, 02h
	out	0f6h, al	; clear a20mask (386+)
	ret

;------------------------------------------------------------------------------
; disables the A20 address line
;	AL - destroyed.

disable_a20:
	test	byte [cs:flags], FLAG_CPU286
	jz	.386

	; for 286
	push	ds
	pushf
	xor	ax, ax
	mov	ds, ax

	push	cs
	push	word .ret_ent
	cli
	mov	[0404h], sp
	mov	[0406h], ss
	mov	al, 0eh
	out	37h, al		; 8255A SHUT0 = 0
	jmp	short $ + 2
	mov	al, 0bh
	out	37h, al		; 8255A SHUT1 = 1
	out	50h, al		; reset NMI
	xor	al, al
	out	0f0h, al	; CPU shutdown
	jmp	short $ + 2

	pop	ds
	pop	ds
.ret_ent:
	mov	al, 0fh
	out	37h, al		; 8255A SHUT0 = 1

	popf
	pop	ds
	ret

.386:
	; for 386+
	mov	al, 03h
	out	0f6h, al	; set a20mask (386+)
	ret

;------------------------------------------------------------------------------
; tests if the A20 address line is enabled.
; compares 256 bytes at 0:0 with ffffh:10h
; Out:	ZF=0 - A20 enabled
;	ZF=1 - A20 disabled

test_a20:
	push	cx
	push	si
	push	di
	push	ds
	push	es

	xor	si, si
	mov	ds, si

	mov	di, 0ffffh
	mov	es, di
	mov	di, 10h

	mov	cx, 100h / 2
	rep	cmpsw

	pop	es
	pop	ds
	pop	di
	pop	si
	pop	cx
	ret

;------------------------------------------------------------------------------
; Interrupt handlers
;------------------------------------------------------------------------------

;------------------------------------------------------------------------------
; new INT1Fh handler

int1f_handler:
	pushf
	cmp	ah, 90h				; is it a block move request?
	je	.do
	cmp	ah, 91h				; is it a protected-prog run request?
	je	.do
	popf
	jmp	far [cs:old_int1f]		; jump to old handler

.do:
	call	test_a20			; check if A20 is on or off
	jz	.a20disabled
	mov	byte [cs:.a20state], 1		; preserve state
	jmp	short .call_old_mover
.a20disabled:
	mov	byte [cs:.a20state], 0
.call_old_mover:
	call	far [cs:old_int1f]
	pushf					; save flags for return
	push	ax
	cmp	byte [cs:.a20state], 0		; see if A20 has to be switched
	jz	.disable_it
	call	enable_a20
	jmp	short .move_done
.disable_it:
	call	disable_a20
.move_done:
	pop	ax
	popf
	iret

.a20state	db	0		; keeps A20 state across INT1fh call

;------------------------------------------------------------------------------
; new INT2Fh handler. Catches Func. 4300h+4310h

int2f_handler:
	pushf
	cmp	ax, 4300h			; is it "Installation Check"?
	jne	.driver_address
	mov	al, 80h				; yes, we are installed ;)
	popf
	iret
.driver_address:
	cmp	ax, 4310h			; is it "Get Driver Address"?
	jne	.call_old2f
	push	cs
	pop	es
	mov	bx, xms_dispatcher
	popf
	iret
.call_old2f:
	popf
	jmp	far [cs:old_int2f]		; jump to old handler

;------------------------------------------------------------------------------
; XMS functions
;------------------------------------------------------------------------------

;------------------------------------------------------------------------------
; returns XMS version number
; In:	AH=0
; Out:	AX=XMS version number
;	BX=internal revision number
;	DX=1 if HMA exists, 0 if not

xms_get_version:
	mov	ax, INTERFACE_VER
	mov	bx, DRIVER_VER
	mov	dx, 1				; HMA is always available
	popf
	retf

;------------------------------------------------------------------------------
; requests HMA
; In:	AH=1
;	DX=space needed in HMA (0ffffh if application tries to request HMA)
; Out:	AX=1 if successful
;	AX=0 if not successful
;	  BL=80h -> function not implemented (implemented here ;) )
;	  BL=81h -> VDISK is detected
;	  BL=90h -> HMA does not exist
;	  BL=91h -> HMA already in use
;	  BL=92h -> DX less than HMA_MIN

xms_request_hma:
	cmp	byte [cs:hma_used], 0		; is HMA already used?
	mov	bl, XMS_HMA_IN_USE
	jnz	..@xrh_err
	cmp	dx, [cs:hma_min]		; is request big enough?
	mov	bl, XMS_HMAREQ_TOO_SMALL
	jb	..@xrh_err
	mov	byte [cs:hma_used], 1		; assign HMA to caller
..@xrh_end:
..@xge_a20_end:
	mov	ax, 1
	xor	bl, bl
	popf
	retf
..@xrh_err:
	xor	ax, ax
	popf
	retf

;------------------------------------------------------------------------------
; releases HMA
; In:	AH=2
; Out:	AX=1 if successful
;	AX=0 if not successful
;	  BL=80h -> function not implemented
;	  BL=81h -> VDISK is detected
;	  BL=90h -> HMA doesn't exist
;	  BL=93h -> HMA wasn't allocated

xms_release_hma:
	cmp	byte [cs:hma_used], 0		; is HMA used?
	mov	bl, XMS_HMA_NOT_USED
	jz	..@xrh_err
	mov	byte [cs:hma_used], 0		; now release it
	jmp	short ..@xrh_end

;------------------------------------------------------------------------------
; global A20 address line enable
; In:	AH=3
; Out:	AX=1 if successful
;	AX=0 if not successful
;	  BL=80h -> function is not implemented
;	  BL=81h -> VDISK is detected
;	  BL=82h -> A20 failure

xms_global_enable_a20:
	call	enable_a20			; enable A20
	call	test_a20			; is it really enabled?
	jz	..@xge_a20_err
	jmp	short ..@xge_a20_end
..@xge_a20_err:
	xor	ax, ax
	mov	bl, XMS_A20_FAILURE
	popf
	retf

;------------------------------------------------------------------------------
; global A20 address line disable
; In:	AH=4
; Out:	AX=1 if successful
;	AX=0 if not successful
;	  BL=80h -> function is not implemented
;	  BL=81h -> VDISK is detected
;	  BL=82h -> A20 failure
;	  BL=84h -> A20 still enabled

xms_global_disable_a20:
	call	disable_a20			; disable A20
	call	test_a20			; is it really disabled?
	jnz	..@xge_a20_err
	jmp	short ..@xge_a20_end

;------------------------------------------------------------------------------
; enables A20 locally
; In:	AH=5
; Out:	AX=1 if A20 is enabled, 0 otherwise
;	BL=80h -> function not implemented
;	BL=81h -> VDISK is detected
;	BL=82h -> A20 failure

xms_local_enable_a20:
	inc	word [cs:a20_locks]		; increase lock counter
	call	enable_a20			; enable it
	call	test_a20			; test if it's really enabled
	jz	..@xge_a20_err
	jmp	short ..@xge_a20_end

;------------------------------------------------------------------------------
; disables A20 locally
; In:	AH=6
; Out:	AX=1 if A20 is disabled, 0 otherwise
;	BL=80h -> function not implemented
;	BL=81h -> VDISK is detected
;	BL=82h -> A20 failure

xms_local_disable_a20:
	dec	word [cs:a20_locks]		; decrease lock counter
	jnz	..@xge_a20_end			; disable only if needed
	call	disable_a20			; disable it
	call	test_a20			; test if it's really disabled
	jnz	..@xge_a20_err
	jmp	short ..@xge_a20_end

;------------------------------------------------------------------------------
; returns the state of A20
; In:	AH=7
; Out:	AX=1 if A20 is physically enabled, AX=0 if not
;	BL=00h -> function was successful
;	BL=80h -> function is not implemented
;	BL=81h -> VDISK is detected

xms_query_a20:
	xor	ax, ax			; suppose A20 is disabled
	call	test_a20
	jz	.xqa_a20dis
	mov	ax, 1
.xqa_a20dis:
	xor	bl, bl
	popf
	retf

;------------------------------------------------------------------------------
; searches a/next free XMS memory block
; In:	DS=CS
;	BX - offset of start handle (if search is continued)
;	CX - remaining handles (if search is continued)
; Out:	CY=1 - no free block
;	  BX - offset of end of handle table
;	CY=0 - free block found
;	  BX - offset of free handle
;	  CX - number of remaining handles

xms_find_free_block:
	mov	bx, driver_end		; start at the beginning of the table
	mov	cx, [xms_num_handles]	; check all handles
.find_free_block:
	cmp	byte [bx + xms_handle.used], 0	; is it used?
	jnz	..@xms_find_next_free_block	; yes, go on
	cmp	word [bx + xms_handle.xbase], 0	; assigned memory block or just blank?
	jnz	.found_block			; assigned, return it
..@xms_find_next_free_block:
	add	bx, xms_handle_size	; skip to next handle
	loop	.find_free_block	; check next handle
	stc				; no free block found, error
	ret
.found_block:
..@found_handle:
	clc				; no error, return
	ret

;------------------------------------------------------------------------------
; searches a/next free XMS memory handle
; In:	DS=CS
;	BX - offset of start handle (if search is continued)
;	CX - remaining handles (if search is continued)
; Out:	CY=1 - no free handle
;	  BX - offset of end of handle table
;	CY=0 - free handle found
;	  BX - offset of free handle
;	  CX - number of remaining handles

xms_find_free_handle:
	mov	bx, driver_end		; start at the beginning of the table
	mov	cx, [xms_num_handles]	; check all handles
.find_free_handle:
	cmp	byte [bx + xms_handle.used], 0	; is it used?
	jnz	..@xms_find_next_free_handle	; yes, go on
	cmp	word [bx + xms_handle.xbase], 0	; really blank handle?
	jz	..@found_handle			; found a blank handle
..@xms_find_next_free_handle:
	add	bx, xms_handle_size	; skip to next handle
	loop	.find_free_handle	; check next handle
	stc				; no free block found, error
	ret

;------------------------------------------------------------------------------
; Verifies that a handle is valid and in use.
; In:	DX - handle to verify
; Out:	CY=1 - not a valid handle
;	  BL=0xa2 - XMS_INVALID_HANDLE
;	  AX=0 - Error return.
;	CY=0 - valid handle
;	  AX=modified

xms_handle_valid:
	mov	ax, dx
	sub	ax, driver_end			; Is the handle below start of handles?
	jb	.not_valid

	push	dx
	push	bx
	xor	dx, dx
	mov	bx, xms_handle_size
	div	bx
	test	dx, dx				; Is the handle aligned on a handle boundary?
	pop	bx
	pop	dx
	jnz	.not_valid

	cmp	ax, [cs:xms_num_handles]	; Is the handle number less than the number of handles?
	jae	.not_valid

	; If we come here, the handle is a valid handle
	push	bx
	mov	bx, dx
	cmp	byte [cs:bx + xms_handle.used], 1	; Is the handle in use?
	pop	bx
	jb	.not_valid

	; Handle is valid.
	clc
	ret

.not_valid:
	; Handle is not valid
	mov	bl, XMS_INVALID_HANDLE
	xor	ax, ax
	stc
	ret

;------------------------------------------------------------------------------
; calculates the move address
; In:	BP - handle (0 if DX:AX should be interpreted as seg:ofs value)
;	BX:CX - length
;	DX:AX - offset
; Out:	CY=0 - valid input (handle).
;	  DX:AX - absolute move address
;	CY=1 - not valid handle or invalid length.
;	  AL=XMS_INVALID_DESTINATION_HANDLE - not valid handle.
;	  AL=XMS_INVALID_LENGTH - invalid length.

xms_get_move_addr:
	test	bp, bp			; translate address in DX:AX?
	jnz	.dont_translate
	; convert segment to absolute address
	push	cx
	mov	cx, dx
	shl	cx, 4
	shr	dx, 12
	add	ax, cx
	adc	dx, 0
	pop	cx

	push	ax
	push	dx
	add	ax, cx
	adc	dx, bx
	jc	.invalid_length1

	;; Assert length + offset < 0xffff:ffff + 1 = 0x10fff0
	;  "offset + length" is no good in case offset = 0xfffffffe or so...
	cmp	dx, 10h
	ja	.invalid_length1
	jb	.end
	cmp	ax, 0fff0h
	ja	.invalid_length1

    .end:
	pop	dx
	pop	ax
	clc
	ret

    .invalid_length1:
	pop	dx
	pop	ax
	mov	al, XMS_INVALID_LENGTH
	stc
	ret

.dont_translate:
	xchg	dx, bp		; Now handle is in DX and upper 16 bits of offset in BP.
	push	ax
	call	xms_handle_valid
	pop	ax
	jc	.not_valid_handle

	push	si
	push	cx
	push	bx

	mov	si, dx		; Put handle in SI.
	mov	dx, bp		; Get upper 16 bits of length back.

	push	ax
	push	dx
	;; Assert length + offset <= size.
	;  "offset + length + 1024 - 1" is no good in case offset or length
	;  equal 0xffffffe or something similarly large.
	add	ax, cx
	adc	dx, bx
	jc      .invalid_length2
	add	ax, 1024 - 1
	adc	dx, 0
	jc      .invalid_length2

	; ... shift so dx:ax is size in KiB.
	mov	cx, dx
	shl	cx, 6
	shr	dx, 10
	shr	ax, 10
	or	ax, cx

	test	dx, dx
	jnz	.invalid_length2

	cmp	ax, [cs:si + xms_handle.xsize]
	ja	.invalid_length2

	pop	dx
	pop	ax

	;; Now everything is ok.
	mov	cx, [cs:si + xms_handle.xbase]	; Get block base address.
	mov	bx, cx
	shr	bx, 6
	shl	cx, 10
	add	ax, cx
	adc	dx, bx

	pop	bx
	pop	cx
	pop	si
	clc
	ret

    .not_valid_handle:
	mov	al, XMS_INVALID_DESTINATION_HANDLE
	ret

    .invalid_length2:
	pop	dx
	pop	ax
	pop	bx
	pop	cx
	pop	si
	mov	al, XMS_INVALID_LENGTH
	stc
	ret

;------------------------------------------------------------------------------
; frees an XMS block
; In:	AH=0ah
;	DX=handle to allocated block that should be freed
; Out:	AX=1 if successful
;	AX=0 if not successful
;	  BL=080h -> function not implemented
;	  BL=081h -> VDISK is detected
;	  BL=0a2h -> handle is invalid
;	  BL=0abh -> handle is locked

xms_free_xms:
	call	xms_handle_valid
	jc	.xms_free_not_valid

	push	bx
	push	cx
	push	dx
	push	di
	push	si
	push	ds
	push	cs
	pop	ds

	mov	si, dx				; Get handle into SI.

	xor	dx, dx				; Set to zero for cmp and mov below.

	cmp	byte [si + xms_handle.used], 1	; Is the block locked?
	jne	.locked				; Yes, bail out.

	cmp	[si + xms_handle.xsize], dx	; Is it a zero-length handle?
	je	.zero_handle

.xms_free_loop:
	mov	di, [si + xms_handle.xbase]	; Get base address.
	add	di, [si + xms_handle.xsize]	; Calculate end-address.

	call	xms_find_free_block		; Check free blocks.
	jc	.xms_free_done			; No, was last handle.

.try_concat:
	cmp	di, [bx + xms_handle.xbase]	; Is it adjacent to old block?
	jne	.try_concat_2

	mov	ax, [bx + xms_handle.xsize]	; Concat.
	add	di, ax
	add	[si + xms_handle.xsize], ax

	mov	[bx + xms_handle.xbase], dx	; Blank handle.
	mov	[bx + xms_handle.xsize], dx

	jmp	short .next

.try_concat_2:
	mov	ax, [bx + xms_handle.xbase]	; Is it adjacent to old block?
	add	ax, [bx + xms_handle.xsize]
	cmp	ax, [si + xms_handle.xbase]
	jne	.not_adjacent

	inc	byte [bx + xms_handle.used]	; This one in use temporaily.
	mov	ax, [si + xms_handle.xsize]	; Concat.
	add	[bx + xms_handle.xsize], ax
	mov	[si + xms_handle.xbase], dx	; Blank handle.
	mov	[si + xms_handle.xsize], dx
	mov	[si + xms_handle.used], dl	; Not in use anymore.
	mov	si, bx
	jmp	short .xms_free_loop

.next:
.not_adjacent:
	call	..@xms_find_next_free_block	; See if there are other blks.
	jnc	.try_concat

.xms_free_done:
	mov	[si + xms_handle.used], dl	; Handle isn't used anymore.
	pop	ds
	pop	si
	pop	di
	pop	dx
	pop	cx
	pop	bx
	mov	ax, 1
	xor	bl, bl
	popf
	retf

.zero_handle:
	mov	[si + xms_handle.xbase], dx	; Blank handle.
	jmp	short .xms_free_done

.locked:
	pop	ds
	pop	si
	pop	di
	pop	dx
	pop	cx
	pop	bx
	xor	ax, ax
	mov	bl, XMS_BLOCK_LOCKED

.xms_free_not_valid:
	popf
	retf

;------------------------------------------------------------------------------
; moves an XMS block
; In:	AH=0bh
;	DS:SI=pointer to XMS move structure
; Out:	AX=1 if successful
;	AX=0 if not successful
;	  BL=080h -> function not implemented
;	  BL=081h -> VDISK is detected
;	  BL=082h -> A20 failure
;	  BL=0a3h -> source handle is invalid
;	  BL=0a4h -> source offset is invalid
;	  BL=0a5h -> destination handle is invalid
;	  BL=0a6h -> destination offset is invalid
;	  BL=0a7h -> length is invalid
;	  BL=0a8h -> move has invalid overlap
;	  BL=0a9h -> parity error

xms_move_xms:
	push	ax
	push	bx
	push	cx
	push	dx
	push	di
	push	si
	push	bp
	push	ds
	push	es
	cli					; no interrupts

	call	test_a20			; get A20 state
	pushf					; save it for later
	jnz	.a20_already_enabled	; We don't need to enable it if it's already is enabled.
	call	enable_a20			; now enable it!

.a20_already_enabled:
	mov	cx, [si + xms_move_strc.len]	; get length
	mov	bx, [si + xms_move_strc.len + 2]
	test	cl, 1				; Is the length even?
	jz	.calc_dest_addr
	jmp	.invalid_length

    .calc_dest_addr
	mov	bp, [si + xms_move_strc.dest_handle]
	mov	ax, [si + xms_move_strc.dest_offset]
	mov	dx, [si + xms_move_strc.dest_offset + 2]
	call	xms_get_move_addr		; get move address
	jnc	.calc_src_addr
	jmp	.xms_move_failure_end

    .calc_src_addr:
	push	ax
	push	dx
	mov	bp, [si + xms_move_strc.src_handle]
	mov	ax, [si + xms_move_strc.src_offset]
	mov	dx, [si + xms_move_strc.src_offset + 2]
	call	xms_get_move_addr		; get move address
    	pop	di
    	pop	si
	jnc	.check_len
	jmp	.src_not_valid

	; src  = dx:ax
	; dest = di:si
	; len  = bx:cx

    .check_len:			; Nothing to copy?
	test	cx, cx
	jnz	.check_addr
	test	bx, bx
	jnz	.check_addr
	jmp	.xms_move_ok

    .check_addr:		; src == dest?
	cmp	ax, si
	jne	.move
	cmp	dx, di
	jne	.move
	jmp	.xms_move_ok

.move:
	push	cs
	pop	ds
	test	byte [flags], FLAG_USE_BIOS
	jz	.do_normal_move
	jmp	.do_bios_move
    .do_normal_move:
	test	byte [flags], FLAG_CPU286
	jnz	.do_286_move
	jmp	.386

; for 286
.do_286_move:
	push	cx
	push	bx
	mov	bp, sp

.286_move_loop:
	pop	bx
	pop	cx
	push	cx
	push	bx

	test	bx, bx
	jnz	.286_length_max
	cmp	cx, 1000h
	jbe	.286_length_ok
    .286_length_max:
	xor	bx, bx
	mov	cx, 1000h

.286_length_ok:
	pusha
	push	cs
	pop	ds
	shr	cx, 1
	cli

	; set loadall286_data
	push	cx
	mov	[loadall286_data + loadall286_tbl.ds_desc + desc_cache286.base_low], ax
	mov	[loadall286_data + loadall286_tbl.ds_desc + desc_cache286.base_high], dl
	mov	[loadall286_data + loadall286_tbl.es_desc + desc_cache286.base_low], si
	mov	ax, di
	mov	[loadall286_data + loadall286_tbl.es_desc + desc_cache286.base_high], al
	mov	[loadall286_data + loadall286_tbl.sp], sp

	mov	ax, ss
	mov	[loadall286_data + loadall286_tbl.bx], ax
	mov	[loadall286_data + loadall286_tbl.ss_reg], ax
	shl	ax, 4
	mov	[loadall286_data + loadall286_tbl.ss_desc + desc_cache286.base_low], ax
	mov	ax, ss
	shr	ah, 4
	mov	[loadall286_data + loadall286_tbl.ss_desc + desc_cache286.base_high], ah
	smsw	[loadall286_data + loadall286_tbl.msw]
	sgdt	[loadall286_data + loadall286_tbl.gdt_desc]
	sidt	[loadall286_data + loadall286_tbl.idt_desc]

	; 800h-866h -> loadall286_save
	mov	ax, 80h
	mov	ds, ax
	push	cs
	pop	es
	xor	si, si
	mov	di, loadall286_save
	mov	cx, loadall286_tbl_size / 2
	rep	movsw

	; loadall286_data -> 800h-866h
	push	ds
	push	es
	pop	ds
	pop	es
	mov	si, loadall286_data
	xor	di, di
	mov	cx, loadall286_tbl_size / 2
	rep	movsw

	; loadall286_data -> cpu
	loadall286
..@to_loadall:

	; loadall286_save -> 800h-866h
    .286_restore_loop:
	db	0b8h		; mov ax, ..@code_seg_286
..@code_seg_286	dw	0
	mov	ss, ax
	mov	dx, [ss:si]
	inc	si
	inc	si
	mov	ax, 80h
	mov	ss, ax
	mov	[ss:di], dx
	inc	di
	inc	di
	loop	.286_restore_loop
	mov	ss, bx

	; move data
	xor	si, si
	xor	di, di
	pop	cx
	rep	movsw

	sti
	popa

	add	si, cx			; dest += copied length.
	adc	di, bx
	add	ax, cx			; src += copied length.
	adc	dx, bx

	sub	[bp + 2], cx		; length -= copied length.
	sbb	[bp], bx
	jnz	.to_286_move_loop
	cmp	word [bp + 2], bx ; bx = 0
	jnz	.to_286_move_loop

	pop	bx
	pop	cx
	jmp	.xms_move_ok

    .to_286_move_loop
	jmp	.286_move_loop

; for 386+
.386:
	push	ax
	smsw	ax
	test	al, 1		; Check if we're in protected mode already.
	pop	ax
	jz	.do_pm_move
	jmp	.do_bios_move	; Already in PM.

; move used protected mode
.do_pm_move:
	pushad
	shl	ebx, 16
	mov	bx, cx
	mov	ecx, ebx
	shl	edi, 16
	mov	di, si
	mov	si, dx
	shl	esi, 16
	mov	si, ax

	; src:  esi
	; dest: edi
	; len:  ecx

	lgdt	[gdt32]				; load GDTR
	mov	eax, cr0
	or	al, 1				; set PE bit
	mov	cr0, eax			; shazamm!
	db	0eah				; JMP FAR
	dw	$ + 4, code16idx		; flush IPQ and load CS sel.

	mov	ax, core32idx
	mov	ds, ax
	mov	es, ax

	cmp	esi, edi	; src <= dest?
	ja	.do_move	; No, go ahead.

	;; src <= dest so we must copy from higher addresses to lower in case the ranges overlap.
	std				; High to low.
	lea	esi, [esi + ecx - 4]	; Adjust addresses for dword size.
	lea	edi, [edi + ecx - 4]
	shr	ecx, 2			; Get number of DWORDS to move.
	jnc	.dword_boundary

	mov	ax, [esi + 2]
	mov	[edi + 2], ax
	dec	esi		; We just moved a word so adjust addresses for that.
	dec	esi
	dec	edi
	dec	edi
	jmp	short .dword_boundary

    .do_move:
	shr	ecx, 2				; get number of DWORDS to move
	jnc	.dword_boundary			; is length a DWORD multiple?

	a32 movsw				; no, move first word to adjust

    .dword_boundary:
	rep	a32 movsd			; now move the main block

	mov	eax, cr0
	and	al, ~ 1				; clear PE bit
	mov	cr0, eax			; shazomm!

	db	0eah				; JMP FAR
	dw	$ + 4
..@code_seg_386	dw	0			; flush IPQ and load CS sel.

	popad

.xms_move_ok:
	popf				; get A20 state
	jnz	.a20_was_enabled	; if it was enabled, don't disable
	call	disable_a20		; it was disabled, so restore state

.a20_was_enabled:
	pop	es
	pop	ds
	pop	bp			; restore everything
	pop	si
	pop	di
	pop	dx
	pop	cx
	pop	bx
	pop	ax
	mov	ax, 1				; success
	popf
	retf

; move used bios
.do_bios_move:
	cmp	dx, di
	ja	.bios_ok
	jb	.next1
	cmp	ax, si
	ja	.bios_ok
    .next1:

	push	ax
	push	dx
	add	ax, cx
	adc	dx, bx
	cmp	dx, di
	ja	.illegal_overlap
	jb	.next2
	cmp	ax, si
	ja	.illegal_overlap
    .next2:
	pop	dx
	pop	ax

.bios_ok:
	push	cs
	pop	es

	push	cx
	push	bx
	mov	bp, sp

.bios_move_loop:
	pop	bx
	pop	cx
	push	cx
	push	bx

	test	bx, bx
	jnz	.bios_length_max
	cmp	cx, 1000h	; 0x1000 is a good value according to Tom Ehlert.
	jbe	.bios_length_ok
    .bios_length_max:
	xor	bx, bx
	mov	cx, 1000h

.bios_length_ok:
	pusha
	cli			; Protect bios_gdt for reentrancy.

	;; Fill in source entries.
	mov	[bios_src_low], ax
	mov	[bios_src_middle], dl
	mov	[bios_src_high], dh

	;; Fill in destination entries.
	mov	[bios_dst_low], si
	mov	ax, di
	mov	[bios_dst_middle], al
	mov	[bios_dst_high], ah

	xor	si, si
	xor	di, di

	mov	bx, bios_gdt
	mov	ah, 90h
	clc
	int	1fh

	sti
	popa
	jc	.a20_failure

	add	si, cx			; dest += copied length.
	adc	di, bx
	add	ax, cx			; src += copied length.
	adc	dx, bx

	sub	[bp + 2], cx		; length -= copied length.
	sbb	[bp], bx
	jnz	.bios_move_loop
	cmp	word [bp + 2], bx ; bx = 0
	jnz	.bios_move_loop

	pop	bx
	pop	cx
	jmp	.xms_move_ok

.src_not_valid:
	cmp	al, XMS_INVALID_DESTINATION_HANDLE	; bl set by xms_get_move_addr.
	jne	.invalid_length
	mov	al, XMS_INVALID_SOURCE_HANDLE
	jmp	short .xms_move_failure_end

.a20_failure:
	pop	bx
	pop	cx
	mov	al, XMS_A20_FAILURE
	jmp	short .xms_move_failure_end

.illegal_overlap:
	pop	dx
	pop	ax
	mov	al, XMS_OVERLAP
	jmp	short .xms_move_failure_end

.invalid_length:
	mov	al, XMS_INVALID_LENGTH
	;  Fall through.

.xms_move_failure_end:
	popf			; Saved a20 state.
	jnz	.failure_a20_was_enabled	; if it was enabled, don't disable
	push	ax
	call	disable_a20		; it was disabled, so restore state
	pop	ax

.failure_a20_was_enabled:
	pop	es
	pop	ds
	pop	bp
	pop	si
	pop	di
	pop	dx
	pop	cx
	pop	bx
	mov	bl, al
	pop	ax
	xor	ax, ax
	popf
	retf

;------------------------------------------------------------------------------
; locks an XMS block
; In:	AH=0ch
;	DX=XMS handle to be locked
; Out:	AX=1 if block is locked
;	  DX:BX=32-bit linear address of block
;	AX=0 if not successful
;	  BL=080h -> function not implemented
;	  BL=081h -> VDISK is detected
;	  BL=0a2h -> handle is invalid
;	  BL=0ach -> lock count overflow
;	  BL=0adh -> lock fails

xms_lock_xms:
	call	xms_handle_valid
	jc	.xms_lock_not_valid

	push	bx
	mov	bx, dx
	inc	byte [cs:bx + xms_handle.used]	; Increase lock count.
	jnz	.locked_successful		; go on if no overflow
	dec	byte [cs:bx + xms_handle.used]	; Decrease lock count.
	pop	bx
	xor	ax, ax
	mov	bl, XMS_LOCK_COUNT_OVERFLOW	; overflow, return with error

.xms_lock_not_valid:
	popf
	retf

.locked_successful:
	mov	ax, [cs:bx + xms_handle.xbase]	; Get block base address
	pop	bx
	mov	bx, ax
	shl	bx, 10
	mov	dx, ax
	shr	dx, 6
	mov	ax, 1
	popf
	retf

;------------------------------------------------------------------------------
; unlocks an XMS block
; In:	AH=0dh
;	DX=XMS handle to unlock
; Out:	AX=1 if block is unlocked
;	AX=0 if not successful
;	  BL=080h -> function not implemented
;	  BL=081h -> VDISK is detected
;	  BL=0a2h -> handle is invalid
;	  BL=0aah -> block is not locked

xms_unlock_xms:
	call	xms_handle_valid
	jc	.xms_unlock_not_valid

	push	bx
	mov	bx, dx
	cmp	byte [cs:bx + xms_handle.used], 1 ; Check if block is locked.
	ja	.is_locked			; Go on if true.
	pop	bx
	xor	ax, ax
	mov	bl, XMS_BLOCK_NOT_LOCKED
	popf
	retf
.is_locked:
	dec	byte [cs:bx + xms_handle.used]	; Decrease lock count.
	pop	bx
	mov	ax, 1
	xor	bl, bl
.xms_unlock_not_valid:
	popf
	retf

;------------------------------------------------------------------------------
; reallocates an UMB
; In:	AH=12h
;	BX=new size for UMB in paragraphs
;	DX=segment of UMB to reallocate
; Out:	AX=1 if successful
;	AX=0 if not successful
;	  BL=080h -> function not implemented
;	  BL=0b0h -> no UMB large enough to satisfy request
;	    DX=size of largest UMB in paragraphs
;	  BL=0b2h -> UMB segment is invalid

xms_realloc_umb:		; Not pretty but saves memory.
	;; Fall through.

;------------------------------------------------------------------------------
; requests an UMB block
; In:	AH=10h
;	DX=size of requested memory block in paragraphs
; Out:	AX=1 if successful
;	  BX=segment number of UMB
;	  DX=actual size of the allocated block in paragraphs
;	AX=0 if not successful
;	  DX=size of largest available UMB in paragraphs
;	  BL=080h -> function not implemented
;	  BL=0b0h -> only a smaller UMB are available
;	  BL=0b1h -> no UMBs are available

xms_request_umb:		; Not pretty but saves memory.
	xor	dx, dx		; No UMB memory.
	;; Fall through.

;------------------------------------------------------------------------------
; releases an UMB block
; In:	AH=11h
;	DX=segment of UMB
; Out:	AX=1 if successful
;	AX=0 if not successful
;	  BL=080h -> function not implemented
;	  BL=0b2h -> UMB segment number is invalid

xms_release_umb:
	xor	ax, ax
	mov	bl, XMS_NOT_IMPLEMENTED
	popf
	retf

%include "xmsrsdnt.inc"

;------------------------------------------------------------------------------
; XMS dispatcher
;------------------------------------------------------------------------------

;------------------------------------------------------------------------------
; XMS dispatcher
; In:	AH - function number
; Out:	AX=0 -> function not supported
;	else see appr. routine

xms_table	dw	xms_get_version		; 00h
		dw	xms_request_hma		; 01h
		dw	xms_release_hma		; 02h
		dw	xms_global_enable_a20	; 03h
		dw	xms_global_disable_a20	; 04h
		dw	xms_local_enable_a20	; 05h
		dw	xms_local_disable_a20	; 06h
		dw	xms_query_a20		; 07h
		dw	xms_query_free_xms	; 08h
		dw	xms_alloc_xms		; 09h
		dw	xms_free_xms		; 0ah
		dw	xms_move_xms		; 0bh
		dw	xms_lock_xms		; 0ch
		dw	xms_unlock_xms		; 0dh
		dw	xms_get_handle_info	; 0eh
		dw	xms_realloc_xms		; 0fh
		dw	xms_request_umb		; 10h
		dw	xms_release_umb		; 11h
		dw	xms_realloc_umb		; 12h

xms_dispatcher:
	jmp	short .dispatcher_entry
	nop					;
	nop					; guarantee hookability
	nop					;
.dispatcher_entry:
	pushf					; save flags
	cld

	cmp	ah, 12h				; Is it a supported function?
	ja	.not_supported

.hook_patch:
	jmp	short .hook_ints
.hook_return:
	push	bx
	xor	bh, bh
	mov	bl, ah
	shl	bx, 1

	mov	ax, [cs:xms_table + bx]
	pop	bx
	jmp	ax

.not_supported:
	xor	ax, ax				; everything else fails
	mov	bl, XMS_NOT_IMPLEMENTED
	popf					; and Yury strikes again...
	retf

.hook_ints:
	test	ah, ah				; non-version call?
	jz	.hook_return			; no, don't patch

	push	ax				; Save registers.
	push	bx
	push	cx
	push	dx
	push	ds
	push	es
	push	cs
	pop	ds

	mov	ax, 351fh			; get INT1Fh vector
	int	21h
	mov	[old_int1f + 2], es
	mov	[old_int1f], bx

	mov	ah, 25h				; install own INT1Fh
	mov	dx, int1f_handler
	int	21h

	mov	cx, [xms_num_handles]		; Get number of handles.
	mov	bx, driver_end - xms_handle_size	; Get start of handle table decreased by one handle.
	inc	cx		; Adjust cx accordingly.
	xor	ax, ax		; Zero out AX so it can be use for cmp and mov below.

.find_end_of_handles:
	add	bx, xms_handle_size
	cmp	[bx + xms_handle.xsize], ax
	loopne	.find_end_of_handles
;	jcxz	.end - Not necessary as we know the last entry ends with xsize == 0.

.clear_table:
	mov	[bx + xms_handle.xbase], ax	; Blank handle.
	mov	[bx + xms_handle.xsize], ax	; Block doesn't occupy any space.
	mov	[bx + xms_handle.used], al	; Handle not used nor locked.
	add	bx, xms_handle_size
	loop	.clear_table

.end:
	pop	es			; Restore registers.
	pop	ds
	pop	dx
	pop	cx
	pop	bx
	pop	ax

	mov	word [cs:.hook_patch], 9090h	; insert two NOPs
	jmp	short .hook_return		; and finish it

;------------------------------------------------------------------------------
; mark for the driver end. above has to be the resident part, below the
; transient.

driver_end:

;------------------------------------------------------------------------------
; 16-bit transient code and data. only used once.
;------------------------------------------------------------------------------

;------------------------------------------------------------------------------
; checks if CPU is a 286
; In:	nothing
; Out:	CY=0 - processor is a 286 or higher
;	CY=1 - processor lower than 286
;	AX - destroyed.

check_cpu:
	pushf
	pop	ax
	test	ax, 8000h
	jnz	.186		; <286

	pushf
	pushf
	pop	ax
	or	ah, 40h
	push	ax
	popf
	pushf
	pop	ax
	popf
	and	ah, 0c0h
	cmp	ah, 40h
	je	.386		; 386+
	or	byte [flags], FLAG_CPU286

.386
	clc
	ret

.186:
	stc
	ret

;------------------------------------------------------------------------------
; checks if A20 can be enabled and disabled
; Out:	CF=0 - A20 switching works
;	CF=1 - A20 failure

check_a20:
	call	enable_a20
	call	test_a20			; TEST_A20 should return ZF=0
	jz	.a20failed
	call	disable_a20
	call	test_a20			; TEST_A20 should return ZF=1
	jnz	.a20failed
	clc
	ret
.a20failed:
	stc
	ret

;------------------------------------------------------------------------------
; Saves a string pointed to by ds:si into command_line.
; command_line is truncated to a length of 255 bytes and upper cased.
; In:	DS:SI - pointer to string.
; Out:	AL - destroyed.

copy_command_line:
	push	di
	push	si
	push	cx
	push	es

	; skip progname
.skip_loop:
	lodsb
	test	al, al
	jnz	.skip_loop

	; load arg
	push	cs
	pop	es
	mov	cx, XMS_COMMAND_LINE_LENGTH_MAX
	mov	di, command_line
.loop:
	lodsb
	cmp	al, 0dh
	je	.the_end	; Stop if we copied 0xd - DOZE have some interesting ideas of ending the command line.
	cmp	al, 0ah
	je	.the_end	; Stop if we copied 0xa - DOZE have some interesting ideas of ending the command line.
	test	al, al
	jz	.the_end	; Stop if we did copy nul, else continue.

	cmp	al, 'a'
	jb	.do_move
	cmp	al, 'z'
	ja	.do_move
	;; Must be a lower case letter,
	add	al, 'A' - 'a'	; which now is uppercase.
.do_move:
	stosb
	loop	.loop

.the_end:
	xor	al, al
	stosb

	pop	es
	pop	cx
	pop	si
	pop	di
	ret

;------------------------------------------------------------------------------
; Analyses the contents of command_line and sets variables accordingly.
; In:	Nothing.
; Out:	AX - destroyed.

interpret_command_line:
	push	es
	push	di
	push	si
	push	cx
	push	bx

	push	ds
	pop	es
	;; First we must step over FDXMS.SYS, which we do by scaning for the first space, tab or ^M character.
	mov	si, command_line		; Remember where search started.
	mov	cx, XMS_COMMAND_LINE_LENGTH_MAX
	add	cx, si				; cx is guard.

.next_arg:
	mov	bx, 0d0ah	; BH=^M, BL=^J
	mov	ax, 0920h	; AH=tab, AL=space
	call	eat_characters

	cmp	byte [si], 0
	jne	.next1		; End of string?
	jmp	.done

    .next1:
	cmp	si, cx
	jb	.next2		; Is this necessary?
	jmp	.done

    .next2:
	mov	bx, si		; Remember current position

	;; BIOS argument.
.try_bios:
	mov	di, bios_str
	call	strcmp
	jc	.try_numhandles

	or	byte [flags], FLAG_USE_BIOS
	jmp	short .next_arg

	;; NUMHANDLES argument.
.try_numhandles:
	mov	si, bx
	mov	di, numhandles_str
	call	strcmp
	jc	.bad_arg

	cmp	byte [si], '='
	jne	.numhandles_no_value_given

	inc	si
	call	read_integer
	jc	.numhandles_no_value_given

	cmp	ax, 2
	jae	.numhandles_at_least_two

	mov	ax, numhandles_too_small_str
	call	print_string
	mov	ax, 2

.numhandles_at_least_two:
	cmp	ax, 1024
	jbe	.numhandles_value_ok

	mov	ax, numhandles_too_big_str
	call	print_string
	mov	ax, 1024

.numhandles_value_ok:
	mov	[xms_num_handles], ax
	jmp	short .next_arg

.numhandles_no_value_given:
	mov	ax, numhandles_no_value_str
	call	print_string
	jmp	short .next_arg

	;; Bad argument.
.bad_arg:
	mov	ax, bad_arg_pre
	call	print_string
	mov	si, bx
	mov	di, bad_arg_arg

.bad_arg_loop:
	mov	al, [si]
	mov	[di], al
	test	al, al
	jz	.bad_arg_end
	cmp	al, ' '
	je	.bad_arg_end
	cmp	al, 9		; tab
	je	.bad_arg_end
	cmp	al, 0ah		; lf
	je	.bad_arg_end
	cmp	al, 0dh		; cr
	je	.bad_arg_end
	inc	si
	inc	di
	jmp	short .bad_arg_loop

.bad_arg_end:
	mov	byte [di], 0
	mov	ax, bad_arg_arg
	call	print_string
	mov	ax, bad_arg_post
	call	print_string

	mov	al, [si]
	test	al, al
	jz	.done

	inc	si
	jmp	.next_arg

	;; The end.
.done:
	pop	bx
	pop	cx
	pop	si
	pop	di
	pop	es
	ret

numhandles_str			db	'NUMHANDLES',0
numhandles_too_small_str	db	'Too small argument to NUMHANDLES increased to 2',13,10,0
numhandles_too_big_str		db	'Too big argument to NUMHANDLES clamped down to 1024',13,10,0
numhandles_no_value_str		db	'NUMHANDLES expects an argument (e.g. NUMHANDLES=32)',13,10,0
bios_str			db	'BIOS',0
bad_arg_pre			db	'Ignoring invalid option: ',0
bad_arg_post			db	13,10,0
bad_arg_arg			times 256 db 0

;------------------------------------------------------------------------------
; Reads an integer from a string pointed ot by SI.
; In:	SI - pointers to string.
; Out:	CY=0 - successful conversion.
;	  SI - updated.
;	  AX - integer read.
;	CY=1 - First character wasn't a digit.
;	  AX - destroyed.

read_integer:
	push	dx

	xor	ah, ah
	mov	al, [si]
	test	al, al
	jz	.failure
	cmp	al, '0'
	jb	.failure
	cmp	al, '9'
	ja	.failure
	sub	al, '0'
	mov	dx, ax

	inc	si
	mov	al, [si]

.decimal_loop:
	cmp	al, '0'
	jb	.done
	cmp	al, '9'
	ja	.done
	sub	al, '0'

	push	ax
	mov	ax, dx
	mov	dx, 10
	mul	dx
	test	dx, dx
	pop	dx
	jnz	.overflow
	add	dx, ax
	jc	.overflow

	inc	si
	mov	al, [si]
	test	al, al
	jz	.done
	jmp	short .decimal_loop

.done:
	mov	ax, dx
	pop	dx
	clc
	ret

.overflow:
    .skip_loop:
	inc	si
	mov	al, [si]
	cmp	al, '0'
	jb	.skip_end
	cmp	al, '9'
	jbe	.skip_loop
    .skip_end:
	mov	dx, 0ffffh
	jmp	short .done

.failure:
	pop	dx
	stc
	ret

;------------------------------------------------------------------------------
; Increases SI until a character that don't match contents of AH, AL, BH or BL is found.
; In:	SI - pointer to string.
;	AL - character to step over
;	AH - character to step over, 0x0 to ignore and to ignore BX contents.
;	BL - character to step over, 0x0 to ignore and to ignore BH contents.
;	BH - character to step over, 0x0 to ignore.
; Out:	SI - updated

eat_characters:
.loop:
	cmp	al, [si]
	jne	.try_ah
	inc	si
	jmp	short .loop
.try_ah:
	test	ah, ah
	jz	.done
	cmp	ah, [si]
	jne	.try_bl
	inc	si
	jmp	short .loop
.try_bl:
	test	bl, bl
	jz	.done
	cmp	bl, [si]
	jne	.try_bh
	inc	si
	jmp	short .loop
.try_bh:
	test	bh, bh
	jz	.done
	cmp	bh, [si]
	jne	.done
	inc	si
	jmp	short .loop

.done:
	ret

;------------------------------------------------------------------------------
; Compares two strings up to DI points at nul.
; In:	DI, SI - pointers to strings to compare.
; Out:	CY=0 - strings equal
;	  DI - points at null
;	  SI - points at character where DI is null.
;	  AL - destroyed
;	CY=1 - strings not equal
;	  DI - points at character which diffs
;	  SI - points at character which diffs
;	  AL - destroyed

strcmp:
.loop:
	mov	al, [di]
	test	al, al		; Is it nul?
	jz	.done		; Yes, we have a match!

	cmp	al, [si]
	jne	.failure	; Not equal so no match.
	inc	si
	inc	di
	jmp	short .loop

.failure:
	stc
	ret

.done:
	clc
	ret

;------------------------------------------------------------------------------
; initializes the driver. called only once!
; may modify DI
; In:	ES:DI - pointer to init structure
;	DS - initialized with code segment

init_message		db	13,10,'FreeDOS(98) XMS-Driver version '
			db	DRIVER_VERSION,' - ',DRIVER98_VERSION,13,10
			db	'Copyright (c) 1995, Till Gerken',13,10
			db	'Copyright 2001, 2002, Martin Stroemberg',13,10,'$'

old_dos			db	'XMS needs at least DOS version 3.00.',13,10,'$'
xms_twice		db	'XMS is already installed.',13,10,'$'
no_286			db	'At least a 80286 is required.',13,10,'$'
a20_error		db	'Unable to switch A20 address line.',13,10,'$'
xms_toosmall		db	'Extended memory is too small or not available.',13,10,'$'

xms_mem_found		db	'K bytes XMS memory detected.',13,10,'$'

error_msg		db	"Driver won't be installed."
init_finished		db	13,10,'$'

command_line		times XMS_COMMAND_LINE_LENGTH_MAX db 0
			db	0

initialize:
	pushf
	push	si
	push	ax
	push	bx
	push	cx
	push	dx
	push	bp

	cld

	mov	ah, 9				; first, welcome the user!
	mov	dx, init_message
	int	21h

	mov	ax, 3000h			; get DOS version number
	int	21h
	xchg	ah, al				; convert to BSD
	cmp	ax, 300h			; we need at least 3.00
	jae	.check_xmm
	mov	dx, old_dos
	jmp	.error_exit

.check_xmm:
	mov	ax, 4300h			; check if XMS is already
	int	2fh				; installed
	cmp	al, 80h
	jne	.check_cpu
	mov	dx, xms_twice
	jmp	.error_exit

.check_cpu:
	call	check_cpu			; do we have at least a 286?
	jnc	.check_mem
	mov	dx, no_286
	jmp	.error_exit

.check_mem:
	push	ds
	xor	ax, ax
	mov	ds, ax

	mov	bx, [0594h]	; protected memory size over 16M (1M count)
	test	bx, bx
	jz	.under16
	mov	dx, 043bh
	in	al, dx
	test	al, 4		; f00000-ffffff system used ?
	jz	.under16
	cmp	bx, 64 - 15
	jb	.under64
	mov	bx, 0ffffh
	jmp	short .check_mem_next
    .under64:
	add	bx, 15	; +15M
	shl	bx, 10	; Kbytes
	jmp	short .check_mem_next
    .under16:
	xor	bh, bh
	mov	bl, [0401h]	; protected memory size under 16M (128K count)
	shl	bx, 7	; Kbytes
    .check_mem_next:
	sub	bx, 64	; - HIMEM size
	mov	bp, bx	; save
	jnc	.use_mem
	pop	ds
	mov	dx, xms_toosmall
	jmp	.error_exit

    .use_mem:
	xor	ax, ax
	mov	[0594h], ax
	mov	[0401h], al
	pop	ds

	call	check_a20			; check A20
	jnc	.check_cmd
	mov	dx, a20_error
	jmp	.error_exit

.check_cmd:
	push	ds
	lds	si, [es:di + init_strc.cmd_line]
	call	copy_command_line
	pop	ds

	call	interpret_command_line

	mov	si, driver_end			; SI points to start of XMS handle table.

	mov	ax, bp	; protected memory (Kbytes)

	mov	word [si + xms_handle.xbase], XMS_START	; Init first block and
	mov	[si + xms_handle.xsize], ax		; give it all
							; available memory.
	mov	byte [si + xms_handle.used], 0	; Handle not used or locked.
	add	si, xms_handle_size

	add	ax, 64	; + HIMEM size
	call	print_dec_number
	mov	ah, 9
	mov	dx, xms_mem_found
	int	21h

	mov	word [si + xms_handle.xsize], 0		; Mark end of initialised handles.

	; for 286
	mov	ax, cs
	mov	[..@code_seg_286], ax
	mov	[loadall286_data + loadall286_tbl.cs_reg], ax
	shl	ax, 4
	mov	[loadall286_data + loadall286_tbl.cs_desc + desc_cache286.base_low], ax
	mov	ax, cs
	shr	ah, 4
	mov	[loadall286_data + loadall286_tbl.cs_desc + desc_cache286.base_high], ah

	; for 386
	mov	ax, cs				; setup descriptors
	mov	[..@code_seg_386], ax		; eliminate relocation entry
	mov	cx, 16
	mul	cx
	or	[code16dsc + 2], ax	; limit(low)
	or	[code16dsc + 4], dx	; limit(high)
	add	[gdt32 + 2], ax
	adc	[gdt32 + 4], dx

	push	es
	mov	ax, 352fh			; get INT2Fh vector
	int	21h
	mov	[old_int2f + 2], es
	mov	[old_int2f], bx
	pop	es

	mov	ah, 25h				; install own INT2Fh
	mov	dx, int2f_handler
	int	21h

	mov	[es:di + 2 + init_strc.end_addr], cs	; set end address
	xor	dx, dx
	mov	ax, xms_handle_size
	mul	word [xms_num_handles]
	add	ax, driver_end
	mov	[es:di + init_strc.end_addr], ax
	mov	word [es:di + request_hdr.status], STATUS_OK	; we're alright

	mov	ah, 9
	mov	dx, init_finished
	int	21h
	or	byte [flags], FLAG_INITIALISED		; Mark that we have initialised.
.exit:
	pop	bp
	pop	dx
	pop	cx
	pop	bx
	pop	ax
	pop	si
	popf
	ret

.error_exit:
	mov	[es:di + 2 + init_strc.end_addr], cs		; set end address
	mov	word [es:di + init_strc.end_addr], 0		; now, we don't use
								; any space
	mov	word [es:di + request_hdr.status], STATUS_BAD	; waaaah!
	mov	ah, 9						; print msg
	int	21h
	mov	dx, error_msg
	int	21h
	mov	dx, init_finished
	int	21h
	jmp	short .exit

;------------------------------------------------------------------------------
; Prints a decimal number, using INT21, AH=0x9.
; In:	AX - the number to be printed.
;	DS - initialized with code segment
print_dec_number:
	push	bx
	push	dx
	push	cx
	push	ax

	mov	bx, .dec_num_start
	mov	cx, 10

.loop:
	xor	dx, dx
	div	cx
	add	dl, '0'
	mov	[bx], dl
	test	ax, ax	; Any digits left?
	jz	.print
	dec	bx
	jmp	short .loop

.print:
	mov	dx, bx
	mov	ah, 9
	int	21h

	pop	ax
	pop	cx
	pop	dx
	pop	bx
	ret

		db	'0000'
.dec_num_start	db	'0$'

%include "printstr.inc"

